# Artifact evaluation

This directory accompanies the paper "QuIP: A P4 Quantum Internet Protocol Prototyping Framework"
submitted to the JSAC Special Issue on "The Quantum Internet: principles, protocols, and
architectures".

This README document and the directory in which it is located contain the following:
1. The raw data used to produce the data plots in the paper.
2. Scripts to produce the plots in the paper from the raw data.
3. Scripts and all the parameters needed to exactly reproduce every single simulation run.
4. Scripts to generate new input data, i.e. newly randomly request times (but at the same average
   rate) and different random state seed for the simulator's random number generator.
5. Scripts to run this newly generated input data to produce new raw data.

## Script pre-requisites

In order to run any of the scripts, you must setup a Python environment and install all the
dependent packages. The scripts have been tested with Python 3.8. They will likely work with a newer
version, but in case of issues, please first try installing Python 3.8 and repeat all the steps with
Python 3.8.

Furthermore, the simulator used to produce these results only works on Linux and Mac OS. Therefore,
if using Windows, you will have to use [WSL](https://learn.microsoft.com/en-us/windows/wsl/install)
to run the scripts in this directory.

### Docker (optional)

You may prefer to run the scripts in a `docker` (or `podman`) container. In that case, we recommend
using the `python:3.8` container. You can run it with the following command:

``` sh
docker run -it -v $(pwd):/workspace --workdir /workspace docker.io/library/python:3.8 bash
```

Replace `docker` with `podman` if using `podman` and you may need to append `:z` to the directory
mount, i.e. `-v $(pwd):/workspace:z`, if running on a Linux distribution with SELinux (e.g. Fedora
or RHEL).

### Python environment

Whether you are running in a docker container or not, you need to setup the Python environment and
install all the dependent packages.

If you don't have `virtualenv` installed, install it with:

``` sh
python3 -m pip install virtualenv
```

Then setup a Python virtual environment:

``` sh
if [ ! -d venv ]; then python3 -m virtualenv venv; fi
source venv/bin/activate
```

If you need to choose a particular Python version, such as 3.8, make sure it is installed first,
e.g. by running `python3.8 --help` and then instead of running the above commands run:

``` sh
if [ ! -d venv ]; then python3 -m virtualenv --python=python3.8 venv; fi
source venv/bin/activate
```

And finally install all the packages and their requirements:

``` sh
python -m pip install --extra-index-url=https://quip:n8L0Uob3luM8SY3PMuBf@pypi.netsquid.org -r requirements.txt
```

The underlying simulator, NetSquid, is proprietary with a license allowing for free use for research
and education purposes. However, in order to download the package one would have to register with
their own credentials on their website. Therefore, the credentials above have been provided to allow
for anonymous review.

## 1. Raw data

The raw data used to produce the plots in the paper can be found in the `experiments/hub/results`
directory. They are organised in subdirectories per simulation scenario (a particular combination of
simulation parameters) and within those subdirectories per timestamped iterations. The results
themselves are in `results.json` files located in each iteration subdirectory.

## 2. Produce plots from the raw data

You can produce the exact same plot that is shown in Fig. 8 of the paper by running:

``` sh
python -m experiments.hub.analysis.bsm_units --show
```

This will show the plot on your screen. If you are running in a docker container you may not see
anything or get an error, in which case try the next step.

To reproduce the exact same PDF source image that was used in the paper, run:

``` sh
python -m experiments.hub.analysis.bsm_units
```

And open the PDF file in `experiments/hub/analysis/figs/bsm_units.pdf`.

## 3. Exactly reproduce raw data

The results directory, `experiments/hub/results`, contains several results for several simulation
runs. Every single result can be reproduced exactly (including the random number generator state
seed). To verify this, you can run the following script for any of the simulation results:

``` sh
python -m netsquid_netrunner.runner --scenario-path experiments/hub/results/<scenario>/<iteration>/scenario.yml --results-path results --stdout
```

Replace `<scenario>` and `<iteration>` with any of the available subdirectories. Note that scenarios
with fewer `bsm-units` take less time to run. This will output the results to the directory
`results/<new-iteration>`.

This command will start a runner which will load the configuration files that the above
`scenario.yml` file refers to. It will first create the network indicated by the
`network_config_file` by creating the nodes and connections using the classes indicated in the
`type_config_file`. In our simulations it will always be a hub-and-spoke network with 16 hosts and
single heralding hub in the middle. The number of BSM units is indicated in the scenario directory
name, but it can also be found in the `network_config_file`. The runner will then use the
`protocol_config_file` to load and start the data-plane, control-plane, and user-space protocols on
the nodes. If you inspect the `protocol_config_file` you will notice that the data-plane protocols
are provided with a `program_file_name` which is a BMv2 JSON file compiled from a P4 program written
for the V1Quantum architecture. Our simulations implement the data plane by running the program from
the JSON file. The P4 source files are included in the same directory as the JSON files. For a
reference to the compiler used, please refer to the paper. The runner will then schedule the input
requests as indicated in the `demand_config_file` (the average rate used to produce this demand
sample is indicated in the scenario directory name) and finally run the simulation using the random
state seed and timing parameters included in the `netsquid_config_file`.

In this case, the simulation will run until `sim_time == 2.0`. The simulation cuts off regardless of
whether or not all requests have completed. Therefore, it is normal to see some (or even many for
scenarios where the network is saturated with a high request rate) requests with `*** NO RESULTS
***`.

To verify the files are the same, you will need the `jq` program. Check if you have it already by
running `jq --help` and if not, it can be installed in the container or any Debian based Linux
distribution with (make sure to prefix with `sudo` if needed):

``` sh
apt update
apt -y install jq
```

You can then verify that the results are the same by first formatting the results with `jq` and
using the `diff` command on the result:

``` sh
jq -S . experiments/hub/results/<scenario>/<iteration>/results.json > original_results.json
jq -S . results/<new-iteration>/results.json > new_results.json
diff original_results.json new_results.json
```

Make sure to use the same values for `<scenario>` and `<iteration>` as for the simulation run. The
`<new-iteration>` directory name will depend on the exact time you ran the reproduction command.

If the results are the same, the only difference that should be seen is the different `argv` field
(i.e. the command used to run the simulation). It differs, because the command used to produce the
new results is different from the one that was used to produce the original results.

You can run the same command for every `<scenario>/<iteration>` combination. There is no script
provided to do that automatically as it is not a common use case.

## 4. Generate new input data

The input data into the simulations has some randomness to it, in particular:
1. The requests are injected at an average rate, but the exact times are randomly generated based on
   the average frequency.
2. The random state seed used by the simulator is random if the simulations are run from scratch
   rather than reproducing existing results.

To generate a new set of input data run:

``` sh
python -m experiments.hub.generate
```

This command will create a new directory `experiments/hub/scenarios` within which it will create
subdirectories for 96 (12 input rates * 8 values of BSM units) scenarios. Each subdirectory will
contain the configuration files for that particular scenario. Some configuration files, such as
`netsquid.yml`, `protocol.yml`, and `type.yml` are shared by all scenarios and thus created directly
in the `experiments/hub` directory.

### 5. Generate results for the new input data

Now, you can run the simulations with the new input data.

**CAUTION: these simulations may take several hours to run**

For reference, the original results were produced in WSL on a quad-core hyperthreaded 2.11 GHz Intel
CPU and it took ~6 hours per iteration.

First remove the old results:

``` sh
rm -rf experiments/hub/results
```

Then, to run a single iteration of the new input data run:

``` sh
python -m netsquid_netrunner.experiment --experiment-path experiments/hub
```

The data will be output to `experiments/hub/results`. You can rerun the plotting script from Step 2
to verify that the results look reasonable. However, note that the plot shown in the paper is an
average over 10 iterations and so the results after only one such iteration may look different.

To produce 10 iterations in total you may run:

``` sh
python -m iterate --experiment-type hub --iterations 10
```

If you ran the previous command to run a single iteration, you may use `--iterations 9` instead. The
`iterate.py` script is a small wrapper that will run the `generate` script from step 4 before every
iteration. This is important, because what we want is a new set of input requests for every
iteration, but injected at the same average rate.

Rerunning the plotting from Step 2 should now result in a plot qualitatively identical to that in
the paper.
