"""Connection classes."""

from typing import Dict
from netsquid.components.cchannel import ClassicalChannel
from netsquid.components.qchannel import QuantumChannel
from netsquid.components.models import FibreDelayModel, FibreLossModel
from netsquid.nodes import DirectConnection


class ClassicalConnection(DirectConnection):
    """A classical NetSquid connection.

    Parameters
    ----------
    name
        The name for this connection.
    fibre_delay_model
        The properties for the delay model.
    properties : optional
        Any other properties of the connection.

    """

    def __init__(self, name: str, fibre_delay_model: Dict, **properties):
        super().__init__(
            name=name,
            channel_AtoB=ClassicalChannel(
                name="AtoB",
                models={"delay_model": FibreDelayModel(**fibre_delay_model)},
                **properties,
            ),
            channel_BtoA=ClassicalChannel(
                name="BtoA",
                models={"delay_model": FibreDelayModel(**fibre_delay_model)},
                **properties,
            ),
        )


class QuantumConnection(DirectConnection):
    """A quantum NetSquid connection.

    Parameters
    ----------
    name
        The name for this connection.
    fibre_delay_model
        The properties for the delay model.
    fibre_loss_model
        The properties for the loss model.
    properties : optional
        Any other properties of the connection.

    """

    def __init__(self, name: str, fibre_delay_model: Dict, fibre_loss_model: Dict, **properties):
        fibre_delay_model = fibre_delay_model if fibre_delay_model is not None else {}
        fibre_loss_model = fibre_loss_model if fibre_loss_model is not None else {}

        super().__init__(
            name=name,
            channel_AtoB=QuantumChannel(
                name="AtoB",
                models={
                    "delay_model": FibreDelayModel(**fibre_delay_model),
                    "quantum_loss_model": FibreLossModel(**fibre_loss_model),
                },
                **properties,
            ),
            channel_BtoA=QuantumChannel(
                name="BtoA",
                models={
                    "delay_model": FibreDelayModel(**fibre_delay_model),
                    "quantum_loss_model": FibreLossModel(**fibre_loss_model),
                },
                **properties,
            ),
        )
