"""An application. A single application consists of distributed program running on two nodes."""

import dataclasses
from typing import Any, Dict, Union

from netsquid.nodes import Node
from pydynaa import Entity, EventHandler, EventType


@dataclasses.dataclass
class Request:
    """An application request.

    Parameters
    ----------
    request_id
        The request identifier.
    host0
        First host (source).
    host1
        Second host (remote).
    parameters
        Request parameters.

    """
    request_id: int
    host0: str
    host1: str
    parameters: Dict[str, Union[int, float]]


class Application(Entity):
    """A quantum network application instance.

    Parameters
    ----------
    host0 : netsquid.nodes.Node
        First host (source).
    host1 : netsquid.nodes.Node
        Second host (remote).
    request
        The request that this application instance is handling.
    request_time
        The simulation time at which this request should be issued to the source.

    """

    def __init__(self, host0: Node, host1: Node, request: Request, request_time: float):
        self._host0 = host0
        self._host1 = host1
        self._request = request

        self._wait_once(
            EventHandler(lambda _: self._start()),
            entity=self,
            event=self._schedule_at(request_time, EventType("REQUEST", "Request")),
        )

    def _start(self) -> None:
        self._host0.netrunner_user_space.request(self._request)

    def print_results(self) -> None:
        """Print the final results of this application."""
        print(f"* Request {self._request.request_id} ({self._host0.name}-{self._host1.name})")

        prot0_results = self._host0.netrunner_user_space.results(self._request.request_id)
        prot1_results = self._host1.netrunner_user_space.results(self._request.request_id)

        if (prot0_results is None) or (prot1_results is None):
            print("*** NO RESULTS ***")
            return

        self._host0.netrunner_user_space.print_results(prot0_results, prot1_results)

    def results(self) -> Dict[str, Any]:
        """Return the results as a dictionary.

        Returns
        -------
        :
            Dictionary of simulation results.

        """
        host0_results = self._host0.netrunner_user_space.results(self._request.request_id)
        host1_results = self._host1.netrunner_user_space.results(self._request.request_id)
        return {
            "request": dataclasses.asdict(self._request),
            "host0": dataclasses.asdict(host0_results) if host0_results is not None else None,
            "host1": dataclasses.asdict(host1_results) if host1_results is not None else None,
        }

    @property
    def complete(self) -> bool:
        """Whether this application instance has completed."""
        return (
            (self._host0.netrunner_user_space.results(self._request.request_id) is not None) and
            (self._host1.netrunner_user_space.results(self._request.request_id) is not None)
        )
