"""Abstract base class for the host protocol that the runner expects from the user."""

import abc
import dataclasses
from typing import Type, Optional

from netsquid.protocols import NodeProtocol

from netsquid_netrunner.demand.application import Request


class HostProtocol(abc.ABC, NodeProtocol):
    """The HostProtocol executes programs (single-node instances of a distributed application).

    Parameters
    ----------
    node : `netsquid.nodes.Node`
        The node which is running this protocol.
    name
        The name of the protocol object.

    """

    def __init__(self, node=None, name=None):
        super().__init__(node, name)

    @abc.abstractmethod
    def request(self, request: Request) -> None:
        """Issue a request.

        Parameters
        ----------
        request
            The request to issue.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def results(self, request_id: int) -> Optional[dataclasses.dataclass]:
        """Retrieve the results for a particular request.

        Parameters
        ----------
        request_id
            The request ID for which to retrieve the results.

        Returns
        -------
        :
            The results object or ``None`` if the results are not ready.

        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def print_results(
            cls: Type,
            host0_results: dataclasses.dataclass,
            host1_results: dataclasses.dataclass,
    ) -> None:
        """Print a summary of the results to stdout.

        Note that this is a class method which takes as input the results from both hosts that
        executed the application. Therefore, it can perform a more involved summary of the results
        than only one of the nodes based on its local results would be able to. For example, having
        access to both results allows this function to calculate a QBER.

        Parameters
        ----------
        host0_results
            The results from host0.
        host1_results
            The results from host1.

        """
        raise NotImplementedError
