"""Module for running scenario groups."""

import argparse
import concurrent
from dataclasses import dataclass
import functools
import multiprocessing
import os
from typing import Callable

from netsquid_netrunner.runner import Runner


@dataclass
class _ScenarioConfig:
    """Single scenario configuration."""
    scenario_file: str
    results_path: str


class _ScenarioRunner:
    """Run a single scenario."""

    def __init__(self, runner_factory: Callable, scenario_config: _ScenarioConfig):
        runner = runner_factory(scenario_config.scenario_file, scenario_config.results_path)
        runner.run()


class _ScenarioManager:
    """Delegate a scenario to its own isolated process."""

    def __init__(self, runner_factory: Callable, scenario_config: _ScenarioConfig):
        # We run each simulation in a process that exists only for the duration of its scenario.
        # This is to ensure NetSquid starts with a completely clean slate every time. A process from
        # a process pool does not guarantee that since it persists between jobs. Furthermore, this
        # helps with memory pressure for long running experiments as each process will fully clean
        # up its memory footprint. Experience shows that long-running NetSquid processes slowly
        # accumulate a large memory footprint otherwise.
        process = multiprocessing.Process(
            target=_ScenarioRunner,
            args=(runner_factory, scenario_config),
        )
        process.start()
        process.join()


class Experiment:
    """Experiment responsible for collecting scenarios and running them.

    Parameters
    ----------
    scenario_root_path
        The root path containing all the scenarios. Scenarios must be a file or a directory
        containing a "scenario.yml" file.
    results_root_path
        The root path where the results will be written to.
    runner_factory : optional
        Custom factory for creating runners to run the scenarios.

    """

    def __init__(
            self,
            scenario_root_path: str,
            results_root_path: str,
            runner_factory: Callable = Runner,
    ):
        self._scenario_root_path = scenario_root_path
        self._results_root_path = results_root_path
        self._runner_factory = runner_factory

    def run(self) -> None:
        """Run all scenarios."""
        scenarios = []
        for scenario in os.listdir(self._scenario_root_path):
            scenario_path = os.path.join(self._scenario_root_path, scenario)
            scenario_file = None

            if os.path.isdir(scenario_path):
                # If it's actually a directory the scenario file is inside.
                scenario_file = os.path.join(scenario_path, "scenario.yml")
                if not os.path.isfile(scenario_file):
                    raise ValueError(f"Missing scenario file: {scenario_file}")
                results_path = os.path.join(
                    self._results_root_path,
                    os.path.basename(scenario_path),
                )
            else:
                scenario_file = scenario_path
                subdir = os.path.splitext(os.path.basename(scenario_file))[0]
                results_path = os.path.join(self._results_root_path, subdir)

            scenarios.append(_ScenarioConfig(scenario_file, results_path))

        # Despite the fact that _ScenarioManager spawns its own process, we still use a
        # ProcessPoolExecutor to indicate that we have CPU bound work. ProcessPoolExecutor's
        # defaults are well suited for this use case.
        with concurrent.futures.ProcessPoolExecutor() as ppe:
            ppe.map(functools.partial(_ScenarioManager, self._runner_factory), scenarios)


def __main():  # pragma: no cover
    __parser = argparse.ArgumentParser(description="Run all scenarios in a particular experiment.")

    __parser.add_argument(
        "--experiment-path",
        type=str,
        required=True,
        help="path to experiment directory",
    )

    __args = __parser.parse_args()

    Experiment(
        scenario_root_path=os.path.join(__args.experiment_path, "scenarios"),
        results_root_path=os.path.join(__args.experiment_path, "results"),
    ).run()


if __name__ == "__main__":  # pragma: no cover
    __main()
