"""Demand generator."""

import contextlib
import dataclasses
import enum
from typing import Any, Tuple, Union, Dict, Optional, List
from typing_extensions import Self
import yaml

from netsquid import SECOND
from numpy import random
from scipy.stats import norm, poisson, expon


class _Request:
    _request_id: int
    _host0: str
    _host1: str
    _parameters: Dict[str, Union[int, float]]
    _request_time: float

    def __init__(self, request_id: int, host0: str, host1: str):
        self._request_id = request_id
        self._host0 = host0
        self._host1 = host1
        self._parameters = {}
        self._request_time = 0

    def set_parameters(self, parameters: Dict[str, Union[int, float]]) -> None:
        # pylint: disable=missing-function-docstring
        self._parameters = parameters

    def set_request_time(self, request_time: float) -> None:
        # pylint: disable=missing-function-docstring
        self._request_time = request_time

    def asdict(self) -> Dict[str, Any]:
        # pylint: disable=missing-function-docstring
        return {
            "request_id": self._request_id,
            "host0": self._host0,
            "host1": self._host1,
            "parameters": self._parameters,
            "request_time": self._request_time,
        }


class _DemandMatrix:
    _requests: List[_Request]

    def __init__(self):
        self._requests = []

    def set_requests(self, requests: List[_Request]) -> None:
        # pylint: disable=missing-function-docstring
        self._requests = requests

    def asdict(self) -> Dict[str, Any]:
        # pylint: disable=missing-function-docstring
        return {
            "requests": list(map(lambda r: r.asdict(), self._requests)),
        }


class _ParameterValueType(str, enum.Enum):
    INT = "INT"
    FLOAT = "FLOAT"

    @staticmethod
    def from_input(value_type: Union[type, str]) -> Self:
        # pylint: disable=missing-function-docstring
        if isinstance(value_type, type):
            return _ParameterValueType.from_str(value_type.__name__)
        return _ParameterValueType.from_str(value_type)

    @staticmethod
    def from_str(typestr: str):
        # pylint: disable=missing-function-docstring
        try:
            return _ParameterValueType(typestr.upper())
        except (ValueError, AttributeError) as ex:
            raise ValueError(f"{_ParameterValueType.__name__} must be one of (int, float)") from ex


class _ParameterDistributionType(enum.Enum):
    NORMAL = enum.auto()
    POISSON = enum.auto()
    EXPONENTIAL = enum.auto()
    FIXED = enum.auto()


@dataclasses.dataclass
class _ParameterDistributionNormal:
    mean: Union[int, float]
    stdev: Union[int, float]


@dataclasses.dataclass
class _ParameterDistributionPoisson:
    mean: Union[int, float]


@dataclasses.dataclass
class _ParameterDistributionExponential:
    mean: Union[int, float]


@dataclasses.dataclass
class _ParameterDistributionFixed:
    value: Union[int, float]


class ParameterDistribution:
    """A request parameter value distribution.

    A `ParameterDistribution` object represents a probability distribution from which the concrete
    values of a request parameter will be generated.

    `ParameterDistribution` objects should be created via the provided class methods for each
    probability distribution type: `normal`, `poisson`, `exponential`, and `fixed`. An attempt to
    create a `ParameterDistribution` object via the constructor will raise an exception.

    """
    _value_type: _ParameterValueType
    _distribution_type: _ParameterDistributionType
    _properties: Union[_ParameterDistributionNormal,
                       _ParameterDistributionPoisson,
                       _ParameterDistributionExponential,
                       _ParameterDistributionFixed]

    __create_key = object()

    def __init__(
            self,
            value_type: Union[type, str],
            distribution_type: _ParameterDistributionType,
            properties: Union[_ParameterDistributionNormal,
                              _ParameterDistributionPoisson,
                              _ParameterDistributionExponential,
                              _ParameterDistributionFixed],
            create_key: object = None,
    ):
        if create_key != ParameterDistribution.__create_key:
            raise TypeError(
                f"{self.__class__.__name__}'s constructor is not publicly available; "
                f"please use the provided class methods: normal, poisson, exponential, or fixed"
            )
        self._value_type = _ParameterValueType.from_input(value_type)
        self._distribution_type = distribution_type
        self._properties = properties

    @classmethod
    def normal(
            cls,
            value_type: Union[type, str],
            mean: Union[int, float],
            stdev: Union[int, float],
    ) -> Self:
        """Create a parameter with a normal probability distribution.

        Parameters
        ----------
        value_type
            The type of the values to be created: ``int`` or ``float``.
        mean
            The mean of the normal distribution.
        stdev
            The standard deviation of the normal distribution.

        Returns
        -------
        :
             The `~ParameterDistribution` object.

        """
        return ParameterDistribution(
            value_type=value_type,
            distribution_type=_ParameterDistributionType.NORMAL,
            properties=_ParameterDistributionNormal(mean=mean, stdev=stdev),
            create_key=cls.__create_key,
        )

    @classmethod
    def poisson(cls, value_type: Union[type, str], mean: Union[int, float]) -> Self:
        """Create a parameter with a poisson probability distribution.

        Parameters
        ----------
        value_type
            The type of the values to be created: ``int`` or ``float``.
        mean
            The mean of the poisson distribution.

        Returns
        -------
        :
             The `ParameterDistribution` object.

        """
        return ParameterDistribution(
            value_type=value_type,
            distribution_type=_ParameterDistributionType.POISSON,
            properties=_ParameterDistributionPoisson(mean=mean),
            create_key=cls.__create_key,
        )

    @classmethod
    def exponential(cls, value_type: Union[type, str], mean: Union[int, float]) -> Self:
        """Create a parameter with an exponential probability distribution.

        Parameters
        ----------
        value_type
            The type of the values to be created: ``int`` or ``float``.
        mean
            The mean of the exponential distribution.

        Returns
        -------
        :
             The `ParameterDistribution` object.

        """
        return ParameterDistribution(
            value_type=value_type,
            distribution_type=_ParameterDistributionType.EXPONENTIAL,
            properties=_ParameterDistributionExponential(mean=mean),
            create_key=cls.__create_key,
        )

    @classmethod
    def fixed(cls, value_type: Union[type, str], value: Union[int, float]) -> Self:
        """Create a parameter with a fixed value.

        Parameters
        ----------
        value_type
            The type of the values to be created: ``int`` or ``float``.
        value
            The fixed value.

        Returns
        -------
        :
             The `ParameterDistribution` object.

        """
        return ParameterDistribution(
            value_type=value_type,
            distribution_type=_ParameterDistributionType.FIXED,
            properties=_ParameterDistributionFixed(value=value),
            create_key=cls.__create_key,
        )

    def generate(self, rng: random.Generator, size: int) -> List[Union[int, float]]:
        # pylint: disable=missing-function-docstring
        if self._distribution_type == _ParameterDistributionType.NORMAL:
            self._properties: _ParameterDistributionNormal
            result = norm.rvs(
                size=size,
                loc=self._properties.mean,
                scale=self._properties.stdev,
                random_state=rng,
            )
        elif self._distribution_type == _ParameterDistributionType.POISSON:
            self._properties: _ParameterDistributionPoisson
            result = poisson.rvs(size=size, mu=self._properties.mean, random_state=rng)
        elif self._distribution_type == _ParameterDistributionType.EXPONENTIAL:
            self._properties: _ParameterDistributionExponential
            result = expon.rvs(size=size, scale=self._properties.mean, random_state=rng)
        else:
            assert self._distribution_type == _ParameterDistributionType.FIXED
            self._properties: _ParameterDistributionFixed
            result = [self._properties.value] * size

        if self._value_type == _ParameterValueType.INT:
            result = map(lambda x: int(x + 0.5), result)
        else:
            assert self._value_type == _ParameterValueType.FLOAT
            result = map(float, result)

        return list(result)


class _TimeDistribution:
    _time_zero: Union[int, float]
    _delta_distribution: ParameterDistribution

    def __init__(self):
        self._time_zero = 0
        self._delta_distribution = ParameterDistribution.fixed(
            value_type=_ParameterValueType.INT, value=0
        )

    def set_time_zero(self, time_zero: Union[int, float]) -> None:
        # pylint: disable=missing-function-docstring
        self._time_zero = time_zero

    def set_delta_distribution(self, delta_distribution: ParameterDistribution) -> None:
        # pylint: disable=missing-function-docstring
        self._delta_distribution = delta_distribution

    def set_average_requests_per_second(self, requests_per_second: Union[int, float]) -> None:
        # pylint: disable=missing-function-docstring
        self._delta_distribution = \
            ParameterDistribution.exponential(float, SECOND / requests_per_second)

    def set_fixed_requests_per_second(self, requests_per_second: Union[int, float]) -> None:
        # pylint: disable=missing-function-docstring
        self._delta_distribution = \
            ParameterDistribution.fixed(float, SECOND / requests_per_second)

    def generate(self, rng: random.Generator, size: int) -> List[float]:
        # pylint: disable=missing-function-docstring
        times = [self._time_zero]

        assert size >= 0
        if size == 0:
            return []

        intervals = self._delta_distribution.generate(rng, size-1)
        for idx in range(0, size-1):
            if intervals[idx] < 0:
                raise ValueError(
                    "generated a negative time interval, please check configuration"
                )
            times.append(times[idx] + intervals[idx])

        return times

    def generate_until(self, rng: random.Generator, until: float) -> List[float]:
        # pylint: disable=missing-function-docstring
        times = [self._time_zero]

        while True:
            interval = self._delta_distribution.generate(rng, 1)[0]
            if interval <= 0:
                raise ValueError(
                    "generated a zero or negative time interval, please check configuration"
                )
            next_time = times[-1] + interval
            if next_time > until:
                break
            times.append(next_time)

        return times


class _DemandType(enum.Enum):
    NUMBER_OF_REQUESTS = enum.auto()
    REQUESTS_PER_HOST = enum.auto()
    REQUESTS_UNTIL = enum.auto()


@dataclasses.dataclass
class _DemandNumberOfRequests:
    number_of_requests: int


@dataclasses.dataclass
class _DemandRequestsPerHost:
    requests_per_host: int


@dataclasses.dataclass
class _DemandRequestsUntil:
    requests_until: float


@dataclasses.dataclass
class _HostSet:
    hosts: List[str]
    hosts0: Optional[List[str]]
    hosts1: Optional[List[str]]


class DemandGenerator:
    """A generator to configure and generate demand (a set of requests) for a simulation.

    A demand properties be configured via the available `set_*` methods and the actual set of
    requests can be generated by calling `generate`.

    """
    _type: _DemandType
    _properties: Union[_DemandNumberOfRequests,
                       _DemandRequestsPerHost,
                       _DemandRequestsUntil]
    _request_parameters: Dict[str, ParameterDistribution]
    _request_times: _TimeDistribution

    def __init__(self):
        self.set_number_of_requests(0)
        self._request_parameters = {}
        self._request_times = _TimeDistribution()

    def set_number_of_requests(self, number_of_requests: int) -> Self:
        """Set the generator to generate a fixed number of requests.

        This option is incompatible with `set_requests_per_host` and `set_requests_until`. Calling
        `set_number_of_requests` will erase any settings configured via the incompatible options.

        Parameters
        ----------
        number_of_requests
            The number of requests to generate.

        Returns
        -------
        :
             The `DemandGenerator` object itself to allow for chaining.

        """
        self._type = _DemandType.NUMBER_OF_REQUESTS
        self._properties = _DemandNumberOfRequests(number_of_requests=number_of_requests)
        return self

    def set_requests_per_host(self, requests_per_host: int) -> Self:
        """Set the generator to generate a fixed number of requests per host.

        This option is incompatible with `set_number_of_requests` and `set_requests_until`. Calling
        `set_requests_per_host` will erase any settings configured via the incompatible options.

        Parameters
        ----------
        requests_per_host
            The number of requests to generate per host.

        Returns
        -------
        :
             The `DemandGenerator` object itself to allow for chaining.

        """
        self._type = _DemandType.REQUESTS_PER_HOST
        self._properties = _DemandRequestsPerHost(requests_per_host=requests_per_host)
        return self

    def set_requests_until(self, requests_until: float) -> Self:
        """Set the generator to generate requests until a certain simulation time.

        This option is incompatible with `set_number_of_requests` and `set_requests_per_host`.
        Calling `set_requests_until` will erase any settings configured via the incompatible
        options.

        Parameters
        ----------
        requests_until
            The time until which requests are to be generated.

        Returns
        -------
        :
             The `DemandGenerator` object itself to allow for chaining.

        """
        self._type = _DemandType.REQUESTS_UNTIL
        self._properties = _DemandRequestsUntil(requests_until=requests_until)
        return self

    def set_request_parameter(self, name: str, distribution: ParameterDistribution) -> Self:
        """Set a request parameter to a particular distribution.

        Parameters
        ----------
        distribution
            The probability distribution to set the parameter to.

        Returns
        -------
        :
             The `DemandGenerator` object itself to allow for chaining.

        """
        self._request_parameters[name] = distribution
        return self

    def set_request_time_zero(self, time_zero: float) -> Self:
        """Set the time at which the first request should be created.

        Parameters
        ----------
        time_zero
            The time at which the first request should be created.

        Returns
        -------
        :
             The `DemandGenerator` object itself to allow for chaining.

        """
        self._request_times.set_time_zero(time_zero)
        return self

    def set_request_time_delta_distribution(self, distribution: ParameterDistribution) -> Self:
        """Set the time delta between requests to a particular distribution.

        This option is incompatible with `set_request_time_average_frequency` and
        `set_request_time_fixed_frequency`. Calling `set_request_time_delta_distribution` will erase
        any settings configured via the incompatible options.

        Parameters
        ----------
        distribution
            The probability distribution to set the time delta to.

        Returns
        -------
        :
             The `DemandGenerator` object itself to allow for chaining.

        """
        self._request_times.set_delta_distribution(distribution)
        return self

    def set_request_time_average_frequency(self, requests_per_second: Union[int, float]) -> Self:
        """Set the average request frequency.

        This option is incompatible with `set_request_time_delta_distribution` and
        `set_request_time_fixed_frequency`. Calling `set_request_time_average_frequency` will erase
        any settings configured via the incompatible options.

        Parameters
        ----------
        requests_per_second
            The average number of requests per second.

        Returns
        -------
        :
             The `DemandGenerator` object itself to allow for chaining.

        """
        self._request_times.set_average_requests_per_second(requests_per_second)
        return self

    def set_request_time_fixed_frequency(self, requests_per_second: Union[int, float]) -> Self:
        """Set the fixed request frequency.

        This option is incompatible with `set_request_time_average_frequency` and
        `set_request_time_delta_distribution`. Calling `set_request_time_fixed_frequency` will erase
        any settings configured via the incompatible options.

        Parameters
        ----------
        requests_per_second
            The fixed number of requests per second.

        Returns
        -------
        :
             The `DemandGenerator` object itself to allow for chaining.

        """
        self._request_times.set_fixed_requests_per_second(requests_per_second)
        return self

    def _generate_requests(self, rng: random.Generator, host_set: _HostSet) -> List[_Request]:
        if self._type == _DemandType.NUMBER_OF_REQUESTS:
            requests = self._generate_number_of_requests(rng, host_set)
        elif self._type == _DemandType.REQUESTS_PER_HOST:
            requests = self._generate_requests_per_host(rng, host_set)
        else:
            assert self._type == _DemandType.REQUESTS_UNTIL
            requests = self._generate_requests_until(rng, host_set)
        return requests

    def _generate_number_of_requests(
            self,
            rng: random.Generator,
            host_set: _HostSet,
    ) -> List[_Request]:
        assert self._type == _DemandType.NUMBER_OF_REQUESTS
        self._properties: _DemandNumberOfRequests
        requests = self._generate_requests_number_of_requests(
            rng,
            host_set,
            self._properties.number_of_requests,
        )
        request_parameters = self._generate_request_parameters(rng, len(requests))
        request_times = self._generate_request_times(rng, len(requests))
        self._merge_requests(requests, request_parameters, request_times)
        return requests

    def _generate_requests_per_host(
            self,
            rng: random.Generator,
            host_set: _HostSet,
    ) -> List[_Request]:
        assert self._type == _DemandType.REQUESTS_PER_HOST
        self._properties: _DemandRequestsPerHost
        requests = self._generate_requests_requests_per_host(
            rng,
            host_set,
            self._properties.requests_per_host,
        )
        request_parameters = self._generate_request_parameters(rng, len(requests))
        request_times = self._generate_request_times(rng, len(requests))
        self._merge_requests(requests, request_parameters, request_times)
        return requests

    def _generate_requests_until(
            self,
            rng: random.Generator,
            host_set: _HostSet,
    ) -> List[_Request]:
        assert self._type == _DemandType.REQUESTS_UNTIL
        self._properties: _DemandRequestsUntil
        request_times = self._generate_request_times_until(rng, self._properties.requests_until)
        requests = self._generate_requests_number_of_requests(rng, host_set, len(request_times))
        request_parameters = self._generate_request_parameters(rng, len(request_times))
        self._merge_requests(requests, request_parameters, request_times)
        return requests

    def _generate_requests_number_of_requests(
            self,
            rng: random.Generator,
            host_set: _HostSet,
            size: int,
    ) -> List[_Request]:
        sampling_host_set = self._get_sampling_host_set(host_set)

        requests = []
        for request_counter in range(size):
            h0, h1 = self._sample_hosts(rng, sampling_host_set)
            requests.append(_Request(request_id=request_counter, host0=h0, host1=h1))

        return requests

    def _generate_requests_requests_per_host(
            self,
            rng: random.Generator,
            host_set: _HostSet,
            requests_per_host: int,
    ) -> List[_Request]:
        assert requests_per_host >= 0
        if requests_per_host == 0:
            return []

        requests = []

        sampling_host_set = self._get_sampling_host_set(host_set)
        hosts, hosts0, hosts1 = \
            sampling_host_set.hosts, sampling_host_set.hosts0, sampling_host_set.hosts1
        requests_left = {h: requests_per_host for h in hosts}

        request_counter = 0
        while (len(hosts) > 1) and ((hosts0 is None) or hosts0) and ((hosts1 is None) or hosts1):
            h0, h1 = self._sample_hosts(rng, sampling_host_set)

            requests.append(_Request(request_id=request_counter, host0=h0, host1=h1))
            request_counter += 1

            requests_left[h0] -= 1
            if requests_left[h0] == 0:
                hosts.remove(h0)
                if hosts0 is not None:
                    hosts0.remove(h0)
                if hosts1 is not None:
                    with contextlib.suppress(ValueError):
                        hosts1.remove(h0)

            requests_left[h1] -= 1
            if requests_left[h1] == 0:
                hosts.remove(h1)
                if hosts0 is not None:
                    with contextlib.suppress(ValueError):
                        hosts0.remove(h1)
                if hosts1 is not None:
                    hosts1.remove(h1)

        return requests

    def _generate_request_times(self, rng: random.Generator, size: int) -> List[float]:
        return self._request_times.generate(rng, size)

    def _generate_request_times_until(self, rng: random.Generator, until: float) -> List[float]:
        return self._request_times.generate_until(rng, until)

    def _generate_request_parameters(
            self,
            rng: random.Generator,
            size: int,
    ) -> List[Dict[str, Union[int, float]]]:
        request_parameters = []
        for _ in range(size):
            request_parameters.append({})

        for name, distribution in self._request_parameters.items():
            values = distribution.generate(rng, size)
            for idx, value in enumerate(values):
                request_parameters[idx][name] = value

        return request_parameters

    @staticmethod
    def _merge_requests(
            requests: List[_Request],
            request_parameters: List[Dict[str, Union[int, float]]],
            request_times: List[float],
    ) -> None:
        assert len(requests) == len(request_parameters)
        assert len(requests) == len(request_times)
        for req, time, parameters in zip(requests, request_times, request_parameters):
            req.set_parameters(parameters)
            req.set_request_time(time)

    @staticmethod
    def _get_sampling_host_set(host_set: _HostSet) -> _HostSet:
        hosts = list(host_set.hosts)

        hosts0 = list(host_set.hosts0) if host_set.hosts0 is not None else None
        hosts1 = list(host_set.hosts1) if host_set.hosts1 is not None else None

        return _HostSet(hosts=hosts, hosts0=hosts0, hosts1=hosts1)

    @staticmethod
    def _sample_hosts(rng: random.Generator, host_set: _HostSet) -> Tuple[str, str]:
        hosts, hosts0, hosts1 = host_set.hosts, host_set.hosts0, host_set.hosts1
        h0, h1 = None, None

        if hosts0 is not None:
            h0, = rng.choice(hosts0, 1, replace=False)

        if hosts1 is not None:
            if h0 is not None:
                hosts1 = list(hosts1)
                with contextlib.suppress(ValueError):
                    hosts1.remove(h0)
            h1, = rng.choice(hosts1, 1, replace=False)

        if (h0 is None) and (h1 is None):
            h0, h1 = rng.choice(hosts, 2, replace=False)
        elif h0 is None:
            assert h1 is not None
            hosts0 = list(hosts0) if hosts0 is not None else list(hosts)
            with contextlib.suppress(ValueError):
                hosts0.remove(h1)
            h0, = rng.choice(hosts0, 1, replace=False)
        elif h1 is None:
            assert h0 is not None
            hosts1 = list(hosts1) if hosts1 is not None else list(hosts)
            with contextlib.suppress(ValueError):
                hosts1.remove(h0)
            h1, = rng.choice(hosts1, 1, replace=False)

        return str(h0), str(h1)

    def generate(
            self,
            demand_file: str,
            rng: random.Generator,
            hosts: List[str],
            hosts0: Optional[List[str]] = None,
            hosts1: Optional[List[str]] = None,
    ) -> None:
        """Generate a demand (a set of requests) and save it to a file.

        Parameters
        ----------
        demand_file
            The file to save the demand to.
        rng : numpy.random.Generator
            The random number generator.
        hosts
            The list of hosts between which the demand is to be generated.
        hosts0 : optional
            The list of hosts to be used as host0. If none provided any host may be used as host0.
        hosts1 : optional
            The list of hosts to be used as host1. If none provided any host may be used as host1.

        """
        # pylint: disable=too-many-arguments
        # Reason: this is the main entrypoint function so it needs all the arguments.
        matrix = _DemandMatrix()
        host_set = _HostSet(hosts=hosts, hosts0=hosts0, hosts1=hosts1)
        matrix.set_requests(self._generate_requests(rng, host_set))
        with open(demand_file, "w", encoding="utf-8") as yaml_file:
            yaml.dump(matrix.asdict(), yaml_file, sort_keys=False)
