"""NetSquid configuration generator."""

from typing import Optional
from typing_extensions import Self

import yaml

from netsquid_netrunner.netsquid import netsquid_defaults


class NetsquidGenerator:
    """NetSquid generator."""

    _random_state_seed: Optional[int]
    _time_zero: float
    _time_step: float
    _time_limit: float

    def __init__(self):
        self._random_state_seed = netsquid_defaults.DEFAULT_RANDOM_STATE_SEED
        self._time_zero = netsquid_defaults.DEFAULT_TIME_ZERO
        self._time_step = netsquid_defaults.DEFAULT_TIME_STEP
        self._time_limit = netsquid_defaults.DEFAULT_TIME_LIMIT

    def set_random_state_seed(self, random_state_seed: Optional[int]) -> Self:
        """Set the random state seed for the NetSquid simulation.

        If not set, or set to `None`, a random seed will be generated with every run.

        Parameters
        ----------
        random_state_seed : optional
            The seed value for the NetSquid simulator.

        Returns
        -------
        :
             The `NetsquidGenerator` object itself to allow for chaining.

        """
        if (random_state_seed is not None) and (not isinstance(random_state_seed, int)):
            raise TypeError("random_state_seed must be None or an int")
        self._random_state_seed = random_state_seed
        return self

    def set_time_zero(self, time_zero: float) -> Self:
        """Set the starting time for the NetSquid simulation.

        Parameters
        ----------
        time_zero
            The starting time.

        Returns
        -------
        :
             The `NetsquidGenerator` itself to allow chaining.
        """
        if time_zero < 0:
            raise ValueError("NetsquidGenerator time_zero cannot be negative")
        self._time_zero = time_zero
        return self

    def set_time_step(self, time_step: float) -> Self:
        """Set the time step for the NetSquid simulation.

        The has no effect on simulation granularity. The time step is simply the length of
        simulation time after which NetSquid will yield to do some intermediate processing (e.g.
        check and print the status of running protocols).

        Parameters
        ----------
        time_step
            The time step.

        Returns
        -------
        :
             The `NetsquidGenerator` itself to allow chaining.

        """
        if time_step < 0:
            raise ValueError("NetsquidGenerator time_step cannot be negative")
        self._time_step = time_step
        return self

    def set_time_limit(self, time_limit: float) -> Self:
        """Set the time limit for the NetSquid simulation.

        The time limit is the simulation time beyond which NetSquid will not advance anymore.

        Parameters
        ----------
        time_limit
            The time limit at which to end the simulation regardless of status.

        Returns
        -------
        :
             The `NetsquidGenerator` itself to allow chaining.

        """
        if time_limit < 0:
            raise ValueError("NetsquidGenerator time_limit cannot be negative")
        self._time_limit = time_limit
        return self

    def generate(self, netsquid_file: str):
        """Save the NetSquid configuration to a file.

        Parameters
        ----------
        netsquid_file
            The file to save the configuration to.

        """
        config = {
            "random_state_seed": self._random_state_seed,
            "time_zero": self._time_zero,
            "time_step": self._time_step,
            "time_limit": self._time_limit,
        }
        with open(netsquid_file, "w", encoding="utf-8") as yaml_file:
            yaml.dump(config, yaml_file, sort_keys=False)
