"""Network generator."""

import copy
from collections import defaultdict
from dataclasses import dataclass
import enum
from typing import Dict, Optional, Any, List
from typing_extensions import Self
import yaml


def _merge(orig: Dict[Any, Any], custom: Dict[Any, Any]) -> None:
    for key in custom.keys():
        if key in orig:
            if isinstance(orig[key], dict) and isinstance(custom[key], dict):
                _merge(orig[key], custom[key])
            else:
                # We update the orig dict with the value in custom
                orig[key] = custom[key]
        else:
            orig[key] = custom[key]


class NetworkBase:
    """Default network component configuration."""
    _controller: Dict[str, Any]
    _host: Dict[str, Any]
    _heralding_station: Dict[str, Any]
    _repeater: Dict[str, Any]
    _router: Dict[str, Any]
    _classical_connection: Dict[str, Any]
    _quantum_connection: Dict[str, Any]

    def __init__(self):
        self._controller = {}
        self._host = {}
        self._heralding_station = {}
        self._repeater = {}
        self._router = {}
        self._classical_connection = {}
        self._quantum_connection = {}

    @property
    def controller(self) -> Dict[str, Any]:
        """Controller default properties."""
        return self._controller

    @property
    def host(self) -> Dict[str, Any]:
        """Host default properties."""
        return self._host

    @property
    def heralding_station(self) -> Dict[str, Any]:
        """Heralding station default properties."""
        return self._heralding_station

    @property
    def repeater(self) -> Dict[str, Any]:
        """Repeater default properties."""
        return self._repeater

    @property
    def router(self) -> Dict[str, Any]:
        """Router default properties."""
        return self._router

    @property
    def classical_connection(self) -> Dict[str, Any]:
        """Classical connection default properties."""
        return self._classical_connection

    @property
    def quantum_connection(self) -> Dict[str, Any]:
        """Quantum connection default properties."""
        return self._quantum_connection

    def update_controller(self, properties: Dict[str, Any]) -> Self:
        """Update controller default properties.

        This will add any new properties and override existing properties. This will not remove any
        existing properties.

        Parameters
        ----------
        properties
            New default properties

        Returns
        -------
        :
            The `NetworkBase` itself for chaining purposes.

        """
        _merge(self._controller, copy.deepcopy(properties))
        return self

    def update_host(self, properties: Dict[str, Any]) -> Self:
        """Update host default properties.

        This will add any new properties and override existing properties. This will not remove any
        existing properties.

        Parameters
        ----------
        properties
            New default properties

        Returns
        -------
        :
            The `NetworkBase` itself for chaining purposes.

        """
        _merge(self._host, copy.deepcopy(properties))
        return self

    def update_heralding_station(self, properties: Dict[str, Any]) -> Self:
        """Update heralding station default properties.

        This will add any new properties and override existing properties. This will not remove any
        existing properties.

        Parameters
        ----------
        properties
            New default properties

        Returns
        -------
        :
            The `NetworkBase` itself for chaining purposes.

        """
        _merge(self._heralding_station, copy.deepcopy(properties))
        return self

    def update_repeater(self, properties: Dict[str, Any]) -> Self:
        """Update repeater default properties.

        This will add any new properties and override existing properties. This will not remove any
        existing properties.

        Parameters
        ----------
        properties
            New default properties

        Returns
        -------
        :
            The `NetworkBase` itself for chaining purposes.

        """
        _merge(self._repeater, copy.deepcopy(properties))
        return self

    def update_router(self, properties: Dict[str, Any]) -> Self:
        """Update router default properties.

        This will add any new properties and override existing properties. This will not remove any
        existing properties.

        Parameters
        ----------
        properties
            New default properties

        Returns
        -------
        :
            The `NetworkBase` itself for chaining purposes.

        """
        _merge(self._router, copy.deepcopy(properties))
        return self

    def update_classical_connection(self, properties: Dict[str, Any]) -> Self:
        """Update classical connection default properties.

        This will add any new properties and override existing properties. This will not remove any
        existing properties.

        Parameters
        ----------
        properties
            New default properties

        Returns
        -------
        :
            The `NetworkBase` itself for chaining purposes.

        """
        _merge(self._classical_connection, copy.deepcopy(properties))
        return self

    def update_quantum_connection(self, properties: Dict[str, Any]) -> Self:
        """Update quantum connection default properties.

        This will add any new properties and override existing properties. This will not remove any
        existing properties.

        Parameters
        ----------
        properties
            New default properties

        Returns
        -------
        :
            The `NetworkBase` itself for chaining purposes.

        """
        _merge(self._quantum_connection, copy.deepcopy(properties))
        return self


class _NodeType(str, enum.Enum):
    """Node type."""
    CONTROLLER = "CONTROLLER"
    HOST = "HOST"
    HERALDING_STATION = "HERALDING_STATION"
    REPEATER = "REPEATER"
    ROUTER = "ROUTER"


class _Node:
    _type: _NodeType
    _port_names: List[str]
    _properties: Dict[str, Any]

    def __init__(self, node_type: _NodeType):
        assert node_type in set(_NodeType)
        self._type = node_type
        self._port_names = []
        self._properties = {}

    def is_controller(self) -> bool:
        # pylint: disable=missing-function-docstring
        return self._type == _NodeType.CONTROLLER

    def is_host(self) -> bool:
        # pylint: disable=missing-function-docstring
        return self._type == _NodeType.HOST

    def is_heralding_station(self) -> bool:
        # pylint: disable=missing-function-docstring
        return self._type == _NodeType.HERALDING_STATION

    def is_repeater(self) -> bool:
        # pylint: disable=missing-function-docstring
        return self._type == _NodeType.REPEATER

    def has_port(self, port_name: str) -> bool:
        # pylint: disable=missing-function-docstring
        return port_name in self._port_names

    def add_port(self, port_name: str) -> Self:
        # pylint: disable=missing-function-docstring
        self._port_names.append(port_name)
        return self

    def update_properties(self, properties: Dict[str, Any]) -> Self:
        # pylint: disable=missing-function-docstring
        _merge(self._properties, copy.deepcopy(properties))
        return self

    def asdict(self) -> Dict[str, Any]:
        # pylint: disable=missing-function-docstring
        return {
            "type": self._type.lower(),
            "properties": {
                "port_names": self._port_names,
                **self._properties,
            }
        }


@dataclass
class LinkPort:
    """A link end-point.

    Parameters
    ----------
    comp
        The component name.
    port
        The port name.

    """
    comp: str
    port: str


class _LinkType(str, enum.Enum):
    CLASSICAL = "CLASSICAL_CONNECTION"
    QUANTUM = "QUANTUM_CONNECTION"

    def asstr(self) -> str:
        # pylint: disable=missing-function-docstring
        return self.lower()


class _Link:
    """A topology link and its properties."""
    _type: _LinkType
    _link_port_1: LinkPort
    _link_port_2: LinkPort
    _properties: Dict[str, Any]

    def __init__(self, link_type: _LinkType, link_port_1: LinkPort, link_port_2: LinkPort):
        self._type = link_type
        self._link_port_1 = link_port_1
        self._link_port_2 = link_port_2
        self._properties = {}

    @property
    def label(self) -> str:
        # pylint: disable=missing-function-docstring
        return (f"{self._link_port_1.comp}_{self._link_port_1.port}_"
                f"{self._link_port_2.comp}_{self._link_port_2.port}")

    def update_properties(self, properties: Dict[str, Any]) -> Self:
        # pylint: disable=missing-function-docstring
        _merge(self._properties, copy.deepcopy(properties))
        return self

    def asdict(self) -> Dict[str, Any]:
        # pylint: disable=missing-function-docstring
        return {
            "type": self._type.asstr(),
            "properties": self._properties,
            "connect_to": {
                "node1": self._link_port_1.comp,
                "node2": self._link_port_2.comp,
                "port_name_node1": self._link_port_1.port,
                "port_name_node2": self._link_port_2.port,
                "label": self.label,
            }
        }


class NetworkGenerator:
    """Create a network topology.

    A network topology is configured by first adding nodes using the `add_*` methods followed by
    connecting them using the `connect_*` methods.

    Parameters
    ----------
    network_base : optional
        The base network configuration.

    """

    def __init__(self, network_base: Optional[NetworkBase] = None):
        self._network_base = network_base if network_base is not None else NetworkBase()
        self._components = {}
        self._quantum_link_count = defaultdict(int)

    def add_controller(self, custom_values: Optional[Dict[str, Any]] = None) -> Self:
        """Add a controller to the topology.

        Parameters
        ----------
        custom_values : optional
            Custom values that deviate from the defaults for the properties dictionary of this node.

        Returns
        -------
        :
             The `NetworkGenerator` object itself to allow for chaining.

        """
        if "controller" in self._components:
            raise ValueError("There can only be one controller")
        self._components["controller"] = _Node(_NodeType.CONTROLLER)\
            .update_properties(self._network_base.controller)\
            .update_properties({} if custom_values is None else custom_values)
        return self

    def add_host(self, name: str, custom_values: Optional[Dict[str, Any]] = None) -> Self:
        """Add a host to the topology.

        Parameters
        ----------
        name
            The name of the host.
        custom_values: optional
            Custom values that deviate from the defaults for the properties dictionary of this node.

        Returns
        -------
        :
             The `NetworkGenerator` object itself to allow for chaining.

        """
        self._components[name] = _Node(_NodeType.HOST)\
            .update_properties(self._network_base.host)\
            .update_properties({} if custom_values is None else custom_values)
        return self

    def add_heralding_station(
            self,
            name: str,
            custom_values: Optional[Dict[str, Any]] = None,
    ) -> Self:
        """Add a heralding station to the topology.

        Parameters
        ----------
        name
            The name of the heralding station.
        custom_values: optional
            Custom values that deviate from the defaults for the properties dictionary of this node.

        Returns
        -------
        :
             The `NetworkGenerator` object itself to allow for chaining.

        """
        self._components[name] = _Node(_NodeType.HERALDING_STATION)\
            .update_properties(self._network_base.heralding_station)\
            .update_properties({} if custom_values is None else custom_values)
        return self

    def add_repeater(self, name: str, custom_values: Optional[Dict[str, Any]] = None) -> Self:
        """Add a repeater to the topology.

        Parameters
        ----------
        name
            The name of the repeater.
        custom_values: optional
            Custom values that deviate from the defaults for the properties dictionary of this node.

        Returns
        -------
        :
             The `NetworkGenerator` object itself to allow for chaining.

        """
        self._components[name] = _Node(_NodeType.REPEATER)\
            .update_properties(self._network_base.repeater)\
            .update_properties({} if custom_values is None else custom_values)
        return self

    def add_router(self, name: str, custom_values: Optional[Dict[str, Any]] = None) -> Self:
        """Add a router to the topology.

        Parameters
        ----------
        name
            The name of the router.
        custom_values: optional
            Custom values that deviate from the defaults for the properties dictionary of this node.

        Returns
        -------
        :
             The `NetworkGenerator` object itself to allow for chaining.

        """
        self._components[name] = _Node(_NodeType.ROUTER)\
            .update_properties(self._network_base.router)\
            .update_properties({} if custom_values is None else custom_values)
        return self

    def _validate_comps_exist(self, link_port_1: LinkPort, link_port_2: LinkPort) -> None:
        # Check that the components exist.
        for comp in (link_port_1.comp, link_port_2.comp):
            if comp not in self._components:
                raise KeyError(f"Component {comp} does not exist in topology")

    def _validate_no_port_collisions(self, link_port_1: LinkPort, link_port_2: LinkPort) -> None:
        # Check that there are no port collisions.
        for comp, port in (
                (link_port_1.comp, link_port_1.port),
                (link_port_2.comp, link_port_2.port)
        ):
            if self._components[comp].has_port(f"{port}"):
                raise KeyError(f"Port {port} is already in use by {comp}")

    def _validate_no_self_connection(self, link_port_1: LinkPort, link_port_2: LinkPort) -> None:
        # pylint: disable=no-self-use
        # Components cannot connect to themselves.
        if link_port_1.comp == link_port_2.comp:
            raise ValueError(f"Cannot connect a component to itself : {link_port_1.comp}")

    def _validate_quantum_link(self, link_port_1: LinkPort, link_port_2: LinkPort) -> None:
        # Do not connect classical components with a quantum link.
        if self._components[link_port_1.comp].is_controller():
            raise TypeError(f"Controller {link_port_1.comp} cannot have a quantum link")

        if self._components[link_port_2.comp].is_controller():
            raise TypeError(f"Controller {link_port_2.comp} cannot have a quantum link")

        # All quantum links must have a heralding station exactly on one end.
        if not (self._components[link_port_1.comp].is_heralding_station() or
                self._components[link_port_2.comp].is_heralding_station()):
            raise TypeError(
                f"Neither {link_port_1.comp} nor {link_port_2.comp} is a heralding station")

        if (self._components[link_port_1.comp].is_heralding_station() and
                self._components[link_port_2.comp].is_heralding_station()):
            raise TypeError(
                f"Both {link_port_1.comp} and {link_port_2.comp} are heralding stations")

    def _validate_number_of_quantum_links(
            self,
            link_port_1: LinkPort,
            link_port_2: LinkPort,
    ) -> None:
        # Components have a restricted number of quantum links based on their type
        for comp in (link_port_1.comp, link_port_2.comp):
            if self._components[comp].is_host() and self._quantum_link_count[comp] == 1:
                raise ValueError(f"Host {comp} cannot have more than one link")
            if self._components[comp].is_repeater() and self._quantum_link_count[comp] == 2:
                raise ValueError(f"Repeater {comp} cannot have more than two links")

    def connect_classical(
            self,
            link_port_1: LinkPort,
            link_port_2: LinkPort,
            properties: Optional[Dict[str, Any]] = None,
    ) -> Self:
        """Connect two nodes with the provided classical link.

        Parameters
        ----------
        link_port_1
            The first link port.
        link_port_2
            The second link port.
        properties : optional
            Custom properties for the connection.

        Returns
        -------
        :
             The `NetworkGenerator` object itself to allow for chaining.

        """
        self._validate_connect(link_port_1, link_port_2, quantum_link=False)
        self._connect(link_port_1, link_port_2, properties, quantum_link=False)
        return self

    def connect_quantum(
            self,
            link_port_1: LinkPort,
            link_port_2: LinkPort,
            properties: Optional[Dict[str, Any]] = None,
    ) -> Self:
        """Connect two nodes with the provided quantum link.

        Parameters
        ----------
        link_port_1
            The first link port.
        link_port_2
            The second link port.
        properties : optional
            Custom properties for the connection.

        Returns
        -------
        :
             The `NetworkGenerator` object itself to allow for chaining.

        """
        self._validate_connect(link_port_1, link_port_2, quantum_link=True)
        self._connect(link_port_1, link_port_2, properties, quantum_link=True)
        return self

    def _validate_connect(
            self,
            link_port_1: LinkPort,
            link_port_2: LinkPort,
            quantum_link: bool,
    ) -> None:
        # Components must first be added to the topology.
        self._validate_comps_exist(link_port_1, link_port_2)

        # Make sure the assigned ports don't overlap.
        self._validate_no_port_collisions(link_port_1, link_port_2)

        # Do not allow connecting nodes to themselves.
        self._validate_no_self_connection(link_port_1, link_port_2)

        if quantum_link:
            # Ensure the that a quantum link doesn't connect purely classical components.
            self._validate_quantum_link(link_port_1, link_port_2)

            # Police numbers of links that are allowed.
            self._validate_number_of_quantum_links(link_port_1, link_port_2)

    def _connect(
            self,
            link_port_1: LinkPort,
            link_port_2: LinkPort,
            properties: Optional[Dict[str, Any]],
            quantum_link: bool
    ) -> None:
        if quantum_link:
            link = _Link(_LinkType.QUANTUM, link_port_1, link_port_2)\
                .update_properties(self._network_base.quantum_connection)
            self._quantum_link_count[link_port_1.comp] += 1
            self._quantum_link_count[link_port_2.comp] += 1
        else:
            link = _Link(_LinkType.CLASSICAL, link_port_1, link_port_2)\
                .update_properties(self._network_base.classical_connection)

        link.update_properties({} if properties is None else properties)

        self._components[link_port_1.comp].add_port(link_port_1.port)
        self._components[link_port_2.comp].add_port(link_port_2.port)
        self._components[link.label] = link

    def generate(self, network_file: str, name: Optional[str] = None) -> None:
        """Generate the network configuration file.

        Parameters
        ----------
        network_file
            The file to save the network to.
        name : optional
            The name of the network.

        """
        # pylint: disable=no-self-use
        # Reason: ask pylint why it's not clever enough.
        network = {
            "network": "network" if name is None else name,
            "components": {
                name: component.asdict() for name, component in self._components.items()
            },
        }

        with open(network_file, "w", encoding="utf-8") as yaml_file:
            class Dumper(yaml.Dumper):  # pylint: disable=missing-class-docstring,too-many-ancestors
                def ignore_aliases(self, _):
                    return True
            yaml.dump(network, yaml_file, Dumper=Dumper, sort_keys=False)
