"""Protocol configuration generator."""

import copy
from typing import Any, Dict, Optional, Type, Union
from typing_extensions import Self

import yaml

from netsquid_netrunner.generators.util import _fullname


class _NodeProtocolConfig:
    _protocol_class: Optional[str]
    _kwargs: Dict[str, Any]

    def __init__(self):
        self._protocol_class = None
        self._kwargs = {}

    def set_class(self, protocol_class: Union[Type, str]) -> None:
        # pylint: disable=missing-function-docstring
        if isinstance(protocol_class, str):
            self._protocol_class = protocol_class
        else:
            self._protocol_class = _fullname(protocol_class)

    def set_kwargs(self, kwargs: Dict[str, Any]) -> None:
        # pylint: disable=missing-function-docstring
        self._kwargs = copy.deepcopy(kwargs)

    def asdict(self) -> Dict[str, Any]:
        # pylint: disable=missing-function-docstring
        if self._protocol_class is None:
            return None

        config = {"class": self._protocol_class}
        if self._kwargs:
            config["kwargs"] = self._kwargs
        return config


class _NodeTypeConfig:
    _user_space: _NodeProtocolConfig
    _control_plane: _NodeProtocolConfig
    _data_plane: _NodeProtocolConfig

    def __init__(self):
        self._user_space = _NodeProtocolConfig()
        self._control_plane = _NodeProtocolConfig()
        self._data_plane = _NodeProtocolConfig()

    def set_user_space(self, protocol_class: Union[Type, str], kwargs: Dict[str, Any]) -> None:
        # pylint: disable=missing-function-docstring
        self._user_space.set_class(protocol_class)
        self._user_space.set_kwargs(kwargs)

    def set_control_plane(self, protocol_class: Union[Type, str], kwargs: Dict[str, Any]) -> None:
        # pylint: disable=missing-function-docstring
        self._control_plane.set_class(protocol_class)
        self._control_plane.set_kwargs(kwargs)

    def set_data_plane(self, protocol_class: Union[Type, str], kwargs: Dict[str, Any]) -> None:
        # pylint: disable=missing-function-docstring
        self._data_plane.set_class(protocol_class)
        self._data_plane.set_kwargs(kwargs)

    def asdict(self) -> None:
        # pylint: disable=missing-function-docstring
        user_space_config = self._user_space.asdict()
        control_plane_config = self._control_plane.asdict()
        data_plane_config = self._data_plane.asdict()

        config = {}
        if user_space_config is not None:
            config["user_space"] = user_space_config
        if control_plane_config is not None:
            config["control_plane"] = control_plane_config
        if data_plane_config is not None:
            config["data_plane"] = data_plane_config

        return config


class ProtocolGenerator:
    """Configure the classes to use for different protocols."""

    _controller: _NodeTypeConfig
    _host: _NodeTypeConfig
    _heralding_station: _NodeTypeConfig
    _repeater: _NodeTypeConfig
    _router: _NodeTypeConfig

    def __init__(self):
        self._controller = _NodeTypeConfig()
        self._host = _NodeTypeConfig()
        self._heralding_station = _NodeTypeConfig()
        self._repeater = _NodeTypeConfig()
        self._router = _NodeTypeConfig()

    def set_controller_control_plane(self, protocol_class: Union[Type, str], **kwargs) -> Self:
        """Set the class to use for the controller control plane protocol.

        Parameters
        ----------
        protocol_class
            The class to use for the protocol. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').
        kwargs : optional
            Any keyword arguments to pass to every instance of the protocol. Note that, these should
            only be types that can be represented in a YAML configuration file, i.e. no objects.

        Returns
        -------
        :
             The `ProtocolGenerator` itself to allow chaining.

        """
        self._controller.set_control_plane(protocol_class, kwargs)
        return self

    def set_controller_data_plane(self, protocol_class: Union[Type, str], **kwargs) -> Self:
        """Set the class to use for the controller data plane protocol.

        Parameters
        ----------
        protocol_class
            The class to use for the protocol. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').
        kwargs : optional
            Any keyword arguments to pass to every instance of the protocol. Note that, these should
            only be types that can be represented in a YAML configuration file, i.e. no objects.

        Returns
        -------
        :
             The `ProtocolGenerator` itself to allow chaining.

        """
        self._controller.set_data_plane(protocol_class, kwargs)
        return self

    def set_host_user_space(self, protocol_class: Union[Type, str], **kwargs) -> Self:
        """Set the class to use for the host user space protocol.

        Parameters
        ----------
        protocol_class
            The class to use for the protocol. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').
        kwargs : optional
            Any keyword arguments to pass to every instance of the protocol. Note that, these should
            only be types that can be represented in a YAML configuration file, i.e. no objects.

        Returns
        -------
        :
             The `ProtocolGenerator` itself to allow chaining.

        """
        self._host.set_user_space(protocol_class, kwargs)
        return self

    def set_host_control_plane(self, protocol_class: Union[Type, str], **kwargs) -> Self:
        """Set the class to use for the host control plane protocol.

        Parameters
        ----------
        protocol_class
            The class to use for the protocol. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').
        kwargs : optional
            Any keyword arguments to pass to every instance of the protocol. Note that, these should
            only be types that can be represented in a YAML configuration file, i.e. no objects.

        Returns
        -------
        :
             The `ProtocolGenerator` itself to allow chaining.

        """
        self._host.set_control_plane(protocol_class, kwargs)
        return self

    def set_host_data_plane(self, protocol_class: Union[Type, str], **kwargs) -> Self:
        """Set the class to use for the host data plane protocol.

        Parameters
        ----------
        protocol_class
            The class to use for the protocol. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').
        kwargs : optional
            Any keyword arguments to pass to every instance of the protocol. Note that, these should
            only be types that can be represented in a YAML configuration file, i.e. no objects.

        Returns
        -------
        :
             The `ProtocolGenerator` itself to allow chaining.

        """
        self._host.set_data_plane(protocol_class, kwargs)
        return self

    def set_heralding_station_control_plane(
            self,
            protocol_class: Union[Type, str],
            **kwargs,
    ) -> Self:
        """Set the class to use for the heralding station control plane protocol.

        Parameters
        ----------
        protocol_class
            The class to use for the protocol. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').
        kwargs : optional
            Any keyword arguments to pass to every instance of the protocol. Note that, these should
            only be types that can be represented in a YAML configuration file, i.e. no objects.

        Returns
        -------
        :
             The `ProtocolGenerator` itself to allow chaining.

        """
        self._heralding_station.set_control_plane(protocol_class, kwargs)
        return self

    def set_heralding_station_data_plane(self, protocol_class: Union[Type, str], **kwargs) -> Self:
        """Set the class to use for the heralding station data plane protocol.

        Parameters
        ----------
        protocol_class
            The class to use for the protocol. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').
        kwargs : optional
            Any keyword arguments to pass to every instance of the protocol. Note that, these should
            only be types that can be represented in a YAML configuration file, i.e. no objects.

        Returns
        -------
        :
             The `ProtocolGenerator` itself to allow chaining.

        """
        self._heralding_station.set_data_plane(protocol_class, kwargs)
        return self

    def set_repeater_control_plane(self, protocol_class: Union[Type, str], **kwargs) -> Self:
        """Set the class to use for the repeater control plane protocol.

        Parameters
        ----------
        protocol_class
            The class to use for the protocol. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').
        kwargs : optional
            Any keyword arguments to pass to every instance of the protocol. Note that, these should
            only be types that can be represented in a YAML configuration file, i.e. no objects.

        Returns
        -------
        :
             The `ProtocolGenerator` itself to allow chaining.

        """
        self._repeater.set_control_plane(protocol_class, kwargs)
        return self

    def set_repeater_data_plane(self, protocol_class: Union[Type, str], **kwargs) -> Self:
        """Set the class to use for the repeater data plane protocol.

        Parameters
        ----------
        protocol_class
            The class to use for the protocol. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').
        kwargs : optional
            Any keyword arguments to pass to every instance of the protocol. Note that, these should
            only be types that can be represented in a YAML configuration file, i.e. no objects.

        Returns
        -------
        :
             The `ProtocolGenerator` itself to allow chaining.

        """
        self._repeater.set_data_plane(protocol_class, kwargs)
        return self

    def set_router_control_plane(self, protocol_class: Union[Type, str], **kwargs) -> Self:
        """Set the class to use for the router control plane protocol.

        Parameters
        ----------
        protocol_class
            The class to use for the protocol. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').
        kwargs : optional
            Any keyword arguments to pass to every instance of the protocol. Note that, these should
            only be types that can be represented in a YAML configuration file, i.e. no objects.

        Returns
        -------
        :
             The `ProtocolGenerator` itself to allow chaining.

        """
        self._router.set_control_plane(protocol_class, kwargs)
        return self

    def set_router_data_plane(self, protocol_class: Union[Type, str], **kwargs) -> Self:
        """Set the class to use for the router data plane protocol.

        Parameters
        ----------
        protocol_class
            The class to use for the protocol. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').
        kwargs : optional
            Any keyword arguments to pass to every instance of the protocol. Note that, these should
            only be types that can be represented in a YAML configuration file, i.e. no objects.

        Returns
        -------
        :
             The `ProtocolGenerator` itself to allow chaining.

        """
        self._router.set_data_plane(protocol_class, kwargs)
        return self

    def generate(self, protocol_file: str):
        """Save the protocol configuration to a file.

        Parameters
        ----------
        protocol_file
            The file to save the configuration to.

        """
        config = {
            "controller": self._controller.asdict(),
            "host": self._host.asdict(),
            "heralding_station": self._heralding_station.asdict(),
            "repeater": self._repeater.asdict(),
            "router": self._router.asdict(),
        }
        with open(protocol_file, "w", encoding="utf-8") as yaml_file:
            yaml.dump(config, yaml_file, sort_keys=False)
