"""Type configuration generator."""

from typing import Optional, Type, Union
from typing_extensions import Self

import yaml

from netsquid_netrunner.generators.util import _fullname


class TypeGenerator:
    """Configure the classes to use for different components."""

    _controller: Optional[str]
    _host: Optional[str]
    _heralding_station: Optional[str]
    _repeater: Optional[str]
    _router: Optional[str]
    _classical_connection: Optional[str]
    _quantum_connection: Optional[str]

    def __init__(self):
        self._controller = None
        self._host = None
        self._heralding_station = None
        self._repeater = None
        self._router = None
        self._classical_connection = None
        self._quantum_connection = None

    def set_controller(self, component_class: Union[Type, str]) -> Self:
        """Set the class to use for controller type components.

        Parameters
        ----------
        component_class
            The class to use for the component. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').

        Returns
        -------
        :
             The `TypeGenerator` itself to allow chaining.

        """
        if isinstance(component_class, str):
            self._controller = component_class
        else:
            self._controller = _fullname(component_class)
        return self

    def set_host(self, component_class: Union[Type, str]) -> Self:
        """Set the class to use for host type components.

        Parameters
        ----------
        component_class
            The class to use for the component. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').

        Returns
        -------
        :
             The `TypeGenerator` itself to allow chaining.

        """
        if isinstance(component_class, str):
            self._host = component_class
        else:
            self._host = _fullname(component_class)
        return self

    def set_heralding_station(self, component_class: Union[Type, str]) -> Self:
        """Set the class to use for heralding station type components.

        Parameters
        ----------
        component_class
            The class to use for the component. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').

        Returns
        -------
        :
             The `TypeGenerator` itself to allow chaining.

        """
        if isinstance(component_class, str):
            self._heralding_station = component_class
        else:
            self._heralding_station = _fullname(component_class)
        return self

    def set_repeater(self, component_class: Union[Type, str]) -> Self:
        """Set the class to use for repeater type components.

        Parameters
        ----------
        component_class
            The class to use for the component. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').

        Returns
        -------
        :
             The `TypeGenerator` itself to allow chaining.

        """
        if isinstance(component_class, str):
            self._repeater = component_class
        else:
            self._repeater = _fullname(component_class)
        return self

    def set_router(self, component_class: Union[Type, str]) -> Self:
        """Set the class to use for router type components.

        Parameters
        ----------
        component_class
            The class to use for the component. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').

        Returns
        -------
        :
             The `TypeGenerator` itself to allow chaining.

        """
        if isinstance(component_class, str):
            self._router = component_class
        else:
            self._router = _fullname(component_class)
        return self

    def set_classical_connection(self, component_class: Union[Type, str]) -> Self:
        """Set the class to use for classical connection type components.

        Parameters
        ----------
        component_class
            The class to use for the component. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').

        Returns
        -------
        :
             The `TypeGenerator` itself to allow chaining.

        """
        if isinstance(component_class, str):
            self._classical_connection = component_class
        else:
            self._classical_connection = _fullname(component_class)
        return self

    def set_quantum_connection(self, component_class: Union[Type, str]) -> Self:
        """Set the class to use for quantum connection type components.

        Parameters
        ----------
        component_class
            The class to use for the component. Can be the class itself or the fully qualified name
            of the class as string (i.e. 'module.class').

        Returns
        -------
        :
             The `TypeGenerator` itself to allow chaining.

        """
        if isinstance(component_class, str):
            self._quantum_connection = component_class
        else:
            self._quantum_connection = _fullname(component_class)
        return self

    def generate(self, type_file: str):
        """Save the type configuration to a file.

        Parameters
        ----------
        type_file
            The file to save the configuration to.

        """
        config = {
            "controller": self._controller,
            "host": self._host,
            "heralding_station": self._heralding_station,
            "repeater": self._repeater,
            "router": self._router,
            "classical_connection": self._classical_connection,
            "quantum_connection": self._quantum_connection,
        }
        with open(type_file, "w", encoding="utf-8") as yaml_file:
            yaml.dump(config, yaml_file, sort_keys=False)
