"""Demand loader."""

import logging
from typing import Dict, List
from typing_extensions import Self
from netsquid.nodes import Node
import yaml

import netsquid as ns

from netsquid_netrunner.demand.application import Application, Request


logger = logging.getLogger(__name__)


class DemandLoader:
    """Demand loader."""

    def __init__(self):
        self._apps = None

    def _check_if_load_called(self):
        if self._apps is None:
            raise RuntimeError("DemandLoader.load must be called first")

    @property
    def apps(self) -> List[Application]:
        """List of loaded applications."""
        self._check_if_load_called()
        return self._apps

    def load(self, hosts: Dict[str, Node], demand_file: str) -> Self:
        """Load the host demand and instantiate the applications.

        Parameters
        ----------
        hosts
            The instantiated host nodes.
        demand_file
            File with the demand matrix.

        Returns
        -------
        :
            Self for convenience.

        """
        with open(demand_file, encoding="utf-8") as yaml_file:
            demand_matrix = yaml.safe_load(yaml_file)

        self._apps = []
        for params in demand_matrix["requests"]:
            if params["request_time"] >= ns.sim_time():
                request = Request(
                    request_id=params["request_id"],
                    host0=params["host0"],
                    host1=params["host1"],
                    parameters=params["parameters"],
                )
                self._apps.append(Application(
                    hosts[params["host0"]],
                    hosts[params["host1"]],
                    request,
                    params["request_time"],
                ))
            else:
                logger.warning(
                    "Request with ID %d has a request time (%f) before simulation's start (%f)",
                    params["request_id"], params["request_time"], ns.sim_time(),
                )

        return self
