"""Utilities for netsquid configuration."""

from typing import Optional, Any
from typing_extensions import Self
import yaml

from netsquid_netrunner.netsquid.netsquid_defaults import netsquid_defaults_asdict
from netsquid_netrunner.netsquid.netsquid_runner import NetsquidRunner


class NetsquidLoader:
    """NetSquid loader."""

    def __init__(self):
        self._runner = None

    def _check_if_load_called(self):
        if self._runner is None:
            raise RuntimeError("NetsquidLoader.load must be called first")

    @property
    def runner(self) -> NetsquidRunner:
        """The NetSquid runner."""
        self._check_if_load_called()
        return self._runner

    def load(
            self,
            netsquid_file: Optional[str] = None,
            netsquid_override: Optional[Any] = None,
    ) -> Self:
        """Load the NetSquid configuration from a file.

        Parameters
        ----------
        netsquid_file : optional
            The netsquid configuration file.
        netsquid_override : optional
            Runtime overrides for the configuration. Can be a dict or an object with matching
            attributes.

        Returns
        -------
        :
            Self for convenience.

        """
        config = netsquid_defaults_asdict()

        if netsquid_file is not None:
            with open(netsquid_file, encoding="utf-8") as yaml_file:
                file_config = yaml.safe_load(yaml_file)

                for opt in config:
                    if opt in file_config:
                        config[opt] = file_config[opt]

        if netsquid_override is not None:
            for opt in config:
                if (hasattr(netsquid_override, opt) and
                        (getattr(netsquid_override, opt) is not None)):
                    config[opt] = getattr(netsquid_override, opt)
                elif (isinstance(netsquid_override, dict) and
                      (netsquid_override.get(opt, None) is not None)):
                    config[opt] = netsquid_override[opt]

        self._runner = NetsquidRunner(**config)
        return self
