"""Utilities for network configuration."""

import importlib
from typing import Any, Dict
from typing_extensions import Self
from netsquid.nodes import Node
import yaml

from netsquid_netconf.builder import ComponentBuilder
from netsquid_netconf.netconf import netconf_generator

from netsquid_netrunner.generators.network import _NodeType


class NetworkLoader:
    """Network loader."""

    def __init__(self):
        self._objects = None
        self._config = None
        self._hosts = None

    def _check_if_load_called(self):
        if (self._objects is None) or (self._config is None) or (self._hosts is None):
            raise RuntimeError("NetworkLoader.load must be called first")

    @property
    def objects(self) -> Dict[str, Any]:
        """Network objects."""
        self._check_if_load_called()
        return self._objects

    @property
    def config(self) -> Dict[str, Any]:
        """Network configuration."""
        self._check_if_load_called()
        return self._config

    @property
    def hosts(self) -> Dict[str, Node]:
        """Dict[str, netsquid.nodes.Node]: The dictionary of host objects in the network."""
        self._check_if_load_called()
        return self._hosts

    @staticmethod
    def _add_types_to_builder(type_config_file: str):
        with open(type_config_file, encoding="utf-8") as yaml_file:
            type_config = yaml.safe_load(yaml_file)

        modules = {}
        for type_name, module_class in type_config.items():
            # Get the module first.
            modname, clsname = module_class.rsplit('.', 1)
            if modname not in modules:
                modules[modname] = importlib.import_module(modname)
            mod = modules[modname]

            # Extract the class.
            device_class = getattr(mod, clsname)

            # Add the mapping to the builder.
            ComponentBuilder.add_type(type_name, device_class)

    def _generate_network(self, network_config_file: str):
        generator = netconf_generator(network_config_file)
        self._objects, _ = next(generator)
        # We ignore the config from netconf since we want the raw config, not a modified one.
        with open(network_config_file, encoding="utf-8") as yaml_file:
            self._config = yaml.safe_load(yaml_file)

    def _configure_controller(self):
        controller = self._objects["components"].get("controller", None)
        if controller is not None:
            controller.network_config = self._config
            controller.ether.network_objects = self._objects

    def _extract_hosts(self):
        self._hosts = dict(filter(
            lambda kv: (self._config["components"][kv[0]]["type"] == _NodeType.HOST.lower()),
            self._objects["components"].items()
        ))

    def _start_nodes(self):
        node_types = tuple(ntype.lower() for ntype in _NodeType)
        for object_name, object_config in self._config["components"].items():
            if object_config["type"] in node_types:
                node = self._objects["components"][object_name]
                if (hasattr(node, "netsquid_netrunner_start") and
                        callable(node.netsquid_netrunner_start)):
                    node.netsquid_netrunner_start()

    def load(self, type_config_file: str, network_config_file: str) -> Self:
        """Load the network configuration from a file.

        Parameters
        ----------
        type_config_file
            The file containing type mappings for netsquid_netconf's builder.
        network_config_file
            The network configuration file for netsquid_netconf.

        Returns
        -------
        :
            Self for convenience.

        """
        self._add_types_to_builder(type_config_file)
        self._generate_network(network_config_file)
        self._configure_controller()
        self._extract_hosts()
        self._start_nodes()
        return self
