"""Utilities for protocol configuration."""

import importlib
from typing import Any, Dict, Type
from typing_extensions import Self
import yaml


class ProtocolLoader:
    """Protocol loader."""

    @staticmethod
    def _load_protocol_class(modules: Dict[str, Any], full_class_name: str) -> Any:
        modname, clsname = full_class_name.rsplit('.', 1)
        if modname not in modules:
            modules[modname] = importlib.import_module(modname)
        mod = modules[modname]
        return getattr(mod, clsname)

    @staticmethod
    def _filter_type_components(network_config: Dict[str, Any], ctype: str) -> Dict[str, Any]:
        return filter(
            lambda kv: ctype == kv[1]["type"], network_config["components"].items()
        )

    @staticmethod
    def _bind_protocols(
            network_objects: Dict[str, Any],
            network_plane_name: str,
            mapping: Dict[str, Any],
            components: Dict[str, Any],
            protocol_class: Type,
    ) -> None:
        for name, config in components:
            comp = network_objects["components"][name]
            protocol_attr_name = f"netrunner_{network_plane_name}"
            assert not hasattr(comp, protocol_attr_name)
            kwargs = mapping.get("kwargs", {})
            setattr(comp, protocol_attr_name, protocol_class(comp, config, **kwargs).start())

    def _load_network_plane(
            self,
            network_objects: Dict[str, Any],
            network_config: Dict[str, Any],
            network_plane_name: str,
            protocol_map: Dict[str, Any],
    ) -> None:
        modules = {}
        for ctype, mapping in protocol_map.items():
            # Get the module first.
            protocol_class = self._load_protocol_class(modules, mapping["class"])

            # Get the components to which we want to attach the components.
            components = self._filter_type_components(network_config, ctype)

            # Bind the protocols.
            self._bind_protocols(
                network_objects,
                network_plane_name,
                mapping,
                components,
                protocol_class,
            )

    def load(
            self,
            network_objects: Dict[str, Any],
            network_config: Dict[str, Any],
            protocol_config_file: str,
    ) -> Self:
        """Load and instantiate the protocol configuration.

        Parameters
        ----------
        network_objects
            The instantiated network object to load the protocols to.
        network_config
            The network configuration.
        protocol_config_file
            The protocol config file that contains the data and control plane configurations.

        Returns
        -------
        :
            Self for convenience.

        """
        with open(protocol_config_file, encoding="utf-8") as yaml_file:
            protocol_config = yaml.safe_load(yaml_file)

        data_plane_map = {}
        control_plane_map = {}
        user_space_map = {}

        for device_type, configuration in protocol_config.items():
            if "data_plane" in configuration:
                data_plane_map[device_type] = configuration["data_plane"]
            if "control_plane" in configuration:
                control_plane_map[device_type] = configuration["control_plane"]
            if "user_space" in configuration:
                user_space_map[device_type] = configuration["user_space"]

        self._load_network_plane(
            network_objects,
            network_config,
            "data_plane",
            data_plane_map,
        )
        self._load_network_plane(
            network_objects,
            network_config,
            "control_plane",
            control_plane_map,
        )
        self._load_network_plane(
            network_objects,
            network_config,
            "user_space",
            user_space_map,
        )

        return self
