"""Wrapper around NetSquid to help netrunner provide guarantees on simulation times."""

from typing import Dict, Optional, Any

import netsquid as ns


class NetsquidRunner:
    """Wrapper around NetSquid to run the simulation.

    This class serves to ensure the start time, time stepping, and time limits are adhered to.

    Parameters
    ----------
    random_state_seed
        The seed for the simulations' random number generator.
    time_zero
        The start time of the simulation.
    time_step
        The time step of the simulation, i.e. how long to run before yielding to the caller.
    time_limit
        The end time for the simulation. Can be infinite.

    """

    _random_state_seed: int
    _time_zero: float
    _time_step: float
    _time_limit: float
    _complete: bool

    def __init__(
            self,
            random_state_seed: Optional[int],
            time_zero: float,
            time_step: float,
            time_limit: float,
    ):
        self._random_state_seed = random_state_seed
        if self._random_state_seed is None:
            ns.set_random_state()
            self._random_state_seed = int(ns.get_random_state().randint(0, 2 ** 32))
        ns.set_random_state(seed=self._random_state_seed)

        if time_zero < 0:
            raise ValueError("NetsquidRunner time_zero cannot be negative")
        self._time_zero = time_zero

        self.update_time_step(time_step)
        self.update_time_limit(time_limit)

        self._complete = False

        self.reset()

    def update_time_step(self, time_step: float) -> None:
        """Update the currently used time step.

        Parameters
        ----------
        time_step
            The time step of the simulation, i.e. how long to run before yielding to the caller.

        """
        if time_step < 0:
            raise ValueError("NetsquidRunner time_step cannot be negative")
        self._time_step = time_step

    def update_time_limit(self, time_limit: float) -> None:
        """Update the currently used end time.

        Parameters
        ----------
        time_limit
             The end time for the simulation. Can be infinite.

        """
        if time_limit < 0:
            raise ValueError("NetsquidRunner time_step cannot be negative")
        self._time_limit = time_limit

    def config_asdict(self) -> Dict[str, Any]:
        """Get the current configuration as a dictionary.

        Returns
        -------
        :
            The dictionary mapping the configuration parameters to their current values.

        """
        return {
            "random_state_seed": self._random_state_seed,
            "time_zero": self._time_zero,
            "time_step": self._time_step,
            "time_limit": self._time_limit,
        }

    @property
    def time_in_seconds(self) -> float:
        """The current simulation time in seconds."""
        return ns.sim_time(ns.SECOND)

    @property
    def complete(self) -> bool:
        """Whether the simulation has reached its time limit."""
        return self._complete

    def reset(self) -> None:
        """Reset the simulation state and rewind the clock to the starting time.

        This will not reseed the simulator. To reseed the simulation, instantiate a new runner.

        """
        ns.sim_reset()
        ns.sim_run(end_time=self._time_zero)

    def step(self) -> None:
        """Run the simulation for the duration of a time step.

        If the time step is larger than the time remaining to the time limit, the simulation will
        only run up to the time limit. Consequently, if the time limit has been reached, the
        simulation will not run when `step` is called.

        """
        if self.complete:
            return

        if ns.sim_time() + self._time_step >= self._time_limit:
            ns.sim_run(end_time=self._time_limit)
            self._complete = True
        else:
            ns.sim_run(duration=self._time_step)
