"""NetSquid network simulation runner."""

import argparse
from copy import deepcopy
import datetime
import dataclasses
import json
import logging
import os
import shutil
import sys
import traceback
from typing import Any, Dict, Optional

import git
import yaml

from netsquid_netrunner.loaders.demand import DemandLoader
from netsquid_netrunner.loaders.netsquid import NetsquidLoader
from netsquid_netrunner.loaders.network import NetworkLoader
from netsquid_netrunner.loaders.protocol import ProtocolLoader


@dataclasses.dataclass
class _Loaders:
    """Loader objects."""
    demand: DemandLoader
    netsquid: NetsquidLoader
    network: NetworkLoader
    protocol: ProtocolLoader


class Runner:
    """Network simulation runner.

    A runner manages the state of a single simulation scenario.

    Parameters
    ----------
    scenario_file
        File containing the scenario parameters.
    results_path
        Directory to store the simulation results in.
    netsquid_override : optional
        Override options for netsquid configuration.
    demand_loader : optional
        Loader object for loading application demand. If ``None``,
        `~netsquid_netrunner.loaders.demand.DemandLoader()` will be used.
    netsquid_loader : optional
        Loader object for loading netsquid runner. If ``None``,
        `~netsquid_netrunner.loaders.netsquid.NetsquidLoader()` will be used.
    network_loader : optional
        Loader object for loading the network topology and components. If ``None``,
        `~netsquid_netrunner.loaders.network.NetworkLoader()` will be used.
    protocol_loader : optional
        Loader object for loading the protocols for the network components. If ``None``,
        `~netsquid_netrunner.loaders.protocol.ProtocolLoader()` will be used.

    """

    def __init__(
            self,
            scenario_file: str,
            results_root: str,
            netsquid_override: Optional[Dict[str, Any]] = None,
            demand_loader: Optional[DemandLoader] = None,
            netsquid_loader: Optional[NetsquidLoader] = None,
            network_loader: Optional[NetworkLoader] = None,
            protocol_loader: Optional[ProtocolLoader] = None,
    ):
        # pylint: disable=too-many-arguments
        # Reason: this is the main entrypoint to the runner.
        with open(scenario_file, encoding="utf-8") as yaml_file:
            self._scenario = yaml.safe_load(yaml_file)
        self._config_path = self._scenario.get("config_path", os.path.dirname(scenario_file))

        self._loaders = _Loaders(
            demand=DemandLoader() if demand_loader is None else demand_loader,
            netsquid=NetsquidLoader() if netsquid_loader is None else netsquid_loader,
            network=NetworkLoader() if network_loader is None else network_loader,
            protocol=ProtocolLoader() if protocol_loader is None else protocol_loader,
        )

        self._load_netsquid(netsquid_override)
        self._load_network()
        self._load_protocols()

        self._results_root = results_root
        self._results_path = None

    def _load_netsquid(self, netsquid_override: Optional[Dict[str, Any]]):
        netsquid_file = None
        if "netsquid_config_file" in self._scenario:
            netsquid_file = os.path.join(self._config_path,
                                         self._scenario["netsquid_config_file"])
        self._loaders.netsquid.load(netsquid_file, netsquid_override)

    def _load_network(self):
        type_config_file = os.path.join(self._config_path, self._scenario["type_config_file"])
        network_config_file = os.path.join(
            self._config_path,
            self._scenario["network_config_file"],
        )
        self._loaders.network.load(type_config_file, network_config_file)

    def _load_protocols(self):
        protocol_config_file = os.path.join(
            self._config_path,
            self._scenario["protocol_config_file"],
        )
        self._loaders.protocol.load(
            self._loaders.network.objects,
            self._loaders.network.config,
            protocol_config_file,
        )

    def _load_demand(self):
        demand_file = os.path.join(self._config_path, self._scenario["demand_config_file"])
        self._loaders.demand.load(self._loaders.network.hosts, demand_file)

    def _create_results_path(self):
        collision_counter = 0
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d-%H%M%S")
        while True:
            try:
                dirname = f"{timestamp}-{collision_counter}"
                self._results_path = os.path.join(self._results_root, dirname)
                os.makedirs(self._results_path)
                break
            except FileExistsError:
                collision_counter += 1

    def _run(self, stdout: bool):
        if stdout:
            print(f"=== sim_time: {self._loaders.netsquid.runner.time_in_seconds}")

        apps = self._loaders.demand.apps
        apps_complete = not bool(apps)
        while not (self._loaders.netsquid.runner.complete or apps_complete):
            self._loaders.netsquid.runner.step()

            apps_complete = True
            for app in apps:
                apps_complete = apps_complete and app.complete
                if not apps_complete:
                    break

            if stdout:
                print(f"=== sim_time: {self._loaders.netsquid.runner.time_in_seconds}")
                for node in self._loaders.network.hosts.values():
                    node.netrunner_user_space.print_status()

        if stdout:
            print("=== RESULTS ===")
            for app in apps:
                app.print_results()

    def run(
            self,
            time_step: Optional[float] = None,
            time_limit: Optional[float] = None,
            stdout: bool = False,
    ):
        """Run the simulation scenario.

        Parameters
        ----------
        time_step : optional
            The simulation time (ns) to run for between checking for completeness and printing
            simulation state. By default the currently configured value will be used.
        time_limit : optional
            The upper time limit when to stop the simulation whether or not it reports that it
            completed. Can be ``float("inf")``. By default the currently configured value will be
            used.
        stdout : optional
            Whether to print out simulation state at each time step.

        """
        if time_step is not None:
            self._loaders.netsquid.runner.update_time_step(time_step)
        if time_limit is not None:
            self._loaders.netsquid.runner.update_time_limit(time_limit)

        if self._results_path is not None:
            raise RuntimeError("saving results for multiple runs of one instance is not supported")

        if self._results_root is not None:
            self._create_results_path()
            self._dump_config()

        try:
            self._load_demand()
            self._run(stdout)
            if self._results_path is not None:
                self._dump_results()

        except Exception:
            if self._results_path is not None:
                errorpath = os.path.join(self._results_path, "error.log")
                with open(errorpath, "w", encoding="utf-8") as error_file:
                    traceback.print_exc(file=error_file)
            raise

    def _dump_config(self):
        assert self._results_path is not None

        # Get the pip freeze output
        freezepath = os.path.join(self._results_path, "freeze.txt")
        os.system(f"pip freeze > {freezepath}")

        # Copy the config files over
        results_scenario = deepcopy(self._scenario)
        results_scenario["config_path"] = self._results_path

        for config_file in [
                "type_config_file",
                "network_config_file",
                "protocol_config_file",
                "demand_config_file",
        ]:
            basename = os.path.basename(self._scenario[config_file])
            shutil.copyfile(os.path.join(self._config_path, self._scenario[config_file]),
                            os.path.join(self._results_path, basename))
            results_scenario[config_file] = basename

        # The runner_config_file for the results is created fresh as the options in the scenario's
        # original file may have been overriden from the commandline.
        netsquid_config_file_path = os.path.join(self._results_path, "netsquid.yml")
        with open(netsquid_config_file_path, "w", encoding="utf-8") as yaml_file:
            yaml.dump(self._loaders.netsquid.runner.config_asdict(), yaml_file, sort_keys=False)
        results_scenario["netsquid_config_file"] = os.path.basename(netsquid_config_file_path)

        # Update and dump scenario
        results_scenario_path = os.path.join(self._results_path, "scenario.yml")
        with open(results_scenario_path, "w", encoding="utf-8") as yaml_file:
            yaml.dump(results_scenario, yaml_file, sort_keys=False)

    def _dump_results(self):
        # And finally dump the results in JSON
        results = {}

        try:
            repo = git.Repo(search_parent_directories=True)
            results["gitsha"] = repo.head.object.hexsha
            if repo.is_dirty():
                logging.warning("repo is dirty")
                results["gitsha"] += "-dirty"
        except git.InvalidGitRepositoryError:
            results["gitsha"] = None

        results["argv"] = sys.argv
        results["app_results"] = [app.results() for app in self._loaders.demand.apps]

        results_path = os.path.join(self._results_path, "results.json")
        with open(results_path, "w", encoding="utf-8") as json_file:
            json_file.write(json.dumps(results))


def __main():  # pragma: no cover
    __parser = argparse.ArgumentParser(description="NetSquid simulation scenario runner.")

    __parser.add_argument(
        "--scenario-path",
        type=str,
        required=True,
        help="path to scenario to use for this run",
    )
    __parser.add_argument(
        "--results-path",
        type=str,
        help="path to the results directory",
    )
    __parser.add_argument(
        "--time-zero",
        type=float,
        help="NetSquid start time for the simulation (default: 0)",
    )
    __parser.add_argument(
        "--time-step",
        type=float,
        help="NetSquid time step for checking simulation status (default: 100,000,000 ns)",
    )
    __parser.add_argument(
        "--time-limit",
        type=float,
        help="NetSquid time limit to stop simulation whether it is complete or not (default: inf)",
    )
    __parser.add_argument(
        "--stdout",
        action="store_true",
        default=False,
        help="if set, print status to stdout",
    )
    __parser.add_argument(
        "--log-level",
        type=str,
        choices=["criticial", "error", "warning", "info", "debug"],
        default="warning",
        help="log level to use (default: warning)"
    )

    __args = __parser.parse_args()

    logging.basicConfig(level=__args.log_level.upper())

    __netsquid_override = {
        "time_zero": __args.time_zero,
        "time_step": __args.time_step,
        "time_limit": __args.time_limit,
    }

    __runner = Runner(__args.scenario_path, __args.results_path, __netsquid_override)
    __runner.run(stdout=__args.stdout)


if __name__ == "__main__":  # pragma: no cover
    __main()
